diff --git a/src/director/doveadm-connection.c b/src/director/doveadm-connection.c
index be2c03f2f05830e8b8d970e96330e39fce9ad286..e37ec8099da74ef0eac45b118c72a117ca8f97c0 100644
--- a/src/director/doveadm-connection.c
+++ b/src/director/doveadm-connection.c
@@ -951,6 +951,7 @@ doveadm_connection_ring_sync_timeout(struct doveadm_connection *conn)
 	doveadm_connection_ring_sync_list_move(conn);
 	o_stream_nsend_str(conn->output, "Ring sync timed out\n");
 
+	i_assert(conn->io == NULL);
 	doveadm_connection_set_io(conn);
 	io_set_pending(conn->io);
 
@@ -1169,9 +1170,10 @@ static void doveadm_connections_continue_reset_cmds(void)
 	}
 }
 
-void doveadm_connections_ring_synced(void)
+void doveadm_connections_ring_synced(struct director *dir)
 {
-	while (doveadm_ring_sync_pending_connections != NULL) {
+	while (doveadm_ring_sync_pending_connections != NULL &&
+	       dir->ring_synced) {
 		struct doveadm_connection *conn =
 			doveadm_ring_sync_pending_connections;
 		doveadm_connection_ring_sync_callback_t *callback =
@@ -1179,9 +1181,11 @@ void doveadm_connections_ring_synced(void)
 
 		conn->ring_sync_callback = NULL;
 		doveadm_connection_ring_sync_list_move(conn);
-		doveadm_connection_set_io(conn);
+		if (conn->io == NULL)
+			doveadm_connection_set_io(conn);
 		io_set_pending(conn->io);
 		callback(conn);
 	}
-	doveadm_connections_continue_reset_cmds();
+	if (dir->ring_synced)
+		doveadm_connections_continue_reset_cmds();
 }
diff --git a/src/director/doveadm-connection.h b/src/director/doveadm-connection.h
index 0b55d85612ec5198c7622a536ab4ab057e73e9ab..04904b617c12708fb67d75a3d48d387d80684445 100644
--- a/src/director/doveadm-connection.h
+++ b/src/director/doveadm-connection.h
@@ -8,6 +8,6 @@ doveadm_connection_init(struct director *dir, int fd);
 void doveadm_connections_deinit(void);
 
 void doveadm_connections_kick_callback(struct director *dir);
-void doveadm_connections_ring_synced(void);
+void doveadm_connections_ring_synced(struct director *dir);
 
 #endif
diff --git a/src/director/main.c b/src/director/main.c
index 96cea27864a635894a4c2754aba8915ee6f1ac4f..36f05bae3f20e012dc64d05a9714d7272d76b767 100644
--- a/src/director/main.c
+++ b/src/director/main.c
@@ -253,7 +253,7 @@ static void director_state_changed(struct director *dir)
 
 	if (dir->to_request != NULL && array_count(&new_requests) == 0)
 		timeout_remove(&dir->to_request);
-	doveadm_connections_ring_synced();
+	doveadm_connections_ring_synced(dir);
 }
 
 static void main_preinit(void)
